import aiofiles
import os
from urllib.parse import urljoin
import urllib.request
import lxml.etree
from fish_tool.net.proxy_16yun import async_get
from fish_tool.log_tool import logs, InfoError

html_char__file_char = {  # 支持windows文件系统 也支持压缩成zip文件
    '/': '·┇',
    '\t': '·t',
    '\\': '·┊',
    '?': '·？',
    ' ': '·▅',
    ':': '·▋',
    '*': '·▆',
    '"': "·'",
    '|': '·▎',
    '<': '·〈',
    '>': '·〉',
}
start__file_char = [
    ('https://', '·S'),
    ('http://', '·H'),
]


def root(*paths):
    path = os.path.join(os.path.dirname(__file__), '../../', *paths)
    return os.path.abspath(path)


def html_to_file_name(html_name):
    file_name = urllib.request.unquote(html_name)  # url编码转成中文
    for start, char in start__file_char:
        if file_name.startswith(start):
            file_name = file_name.replace(start, char)

    for html_char, file_char in html_char__file_char.items():
        file_name = file_name.replace(html_char, file_char)
    return file_name


def file_to_html_name(file_name):
    pre = ''
    for start_html, start_char in start__file_char:
        if file_name.startswith(start_char):
            pre = start_html
            file_name = file_name.replace(start_char, '')
    for html_char, file_char in html_char__file_char.items():
        file_name = file_name.replace(file_char, html_char)
    html_name = pre + urllib.request.quote(file_name)  # 中文转成url编码
    return html_name


def _test_file_name():
    assert len(set(html_char__file_char.values())) == len(html_char__file_char)
    for url in [
        'https://www.bilibili.com/video/BV1cA{41}1s7Kz',
        'https://www.ciweimao.com/book_list/weilai<huan>xiang/',
        'http://www.xiaolvji.com/u/aif(i)sh'
    ]:
        file = html_to_file_name(url)
        url2 = file_to_html_name(file)
        if url2 != url:
            logs.brief.info(f'Error {url2} {url}')
        logs.brief.info(f'{file} ===> {url2}')


async def async_read(path, encoding="utf-8"):
    async with aiofiles.open(path, encoding=encoding) as f:
        return await f.read()


async def async_write(txt, path, mod='w', encoding="utf-8"):
    assert mod in ['w', 'a']
    async with aiofiles.open(path, mod, encoding=encoding) as f:
        await f.write(txt)


async def async_write_html(url, path, charset=None, proxy=False):
    resp = await async_get(url, proxy=proxy)
    if charset:
        text = resp.content.decode(charset, "ignore")
        await async_write(text, path, mod='w', encoding=charset)
    else:
        text = resp.text
        await async_write(text, path, mod='w', encoding="utf-8")


class Etree:
    def __init__(self, html, url, node=None):
        """
        :param html: 网页的文本内容
        :param url:  网页地址
        """
        if node is None:
            self.tree = lxml.etree.HTML(html)
        else:
            self.tree = node
        self.raw_url = url

    def url(self, path):
        # path中的键名必须是小写的（不区分大小写）， 而值是区分大小写的
        out = self.txt(path)
        out = urljoin(self.raw_url, out) if out else ''
        return out

    def urls(self, path):
        # path中的键名必须是小写的（不区分大小写）， 而值是区分大小写的
        ts = self.txts(path)
        out = [urljoin(self.raw_url, t) if t else '' for t in ts]
        return out

    @staticmethod
    def trans_txt(txt):
        if isinstance(txt, lxml.etree._Element):
            txt = txt.text
            if txt is None:
                return None
        txt = txt.replace(u'\xa0', u'')  # \xa0 是不间断空白符
        txt = txt.replace(u'\u3000', u'  ')  # \u3000 是全角的空白符
        return txt

    def txt(self, path):
        # path中的键名必须是小写的（不区分大小写）， 而值是区分大小写的
        if not path:
            return ''
        resps = self.tree.xpath(path)
        out = resps[0] if resps else ''
        out = self.trans_txt(out)
        return out

    def txts(self, path):
        # path中的键名必须是小写的（不区分大小写）， 而值是区分大小写的
        nodes = self.tree.xpath(path)
        out = []
        for txt in nodes:
            txt = self.trans_txt(txt)
            if txt is not None:
                out.append(txt)
        return out

    def nodes(self, path):
        # path中的键名必须是小写的（不区分大小写）， 而值是区分大小写的
        trees = self.tree.xpath(path)
        out = [Etree(html='', url=self.raw_url, node=tree) for tree in trees if tree is not None]
        return out

    def __str__(self):
        return lxml.etree.tostring(self.tree).decode('utf8')

    def __getitem__(self, item):
        return self.tree.attrib[item]

    def __setitem__(self, key, value):
        self.tree.attrib[key] = value



from tqdm import tqdm
from fish_tool.log_tool import logs, InfoError

try:
    import py2neo
except ImportError:
    py2neo = ImportError('未载入 py2neo')
    logs.tmp.warning(f'未载入模块 py2neo')


# a = Node('测试-人', name='张三')
# b = Node('测试-人', name='李四', 性别='男')
# # r = Relationship(a, '认识', b)
# print(a)
# print(b)


def node_str_fn(self):
    label = ','.join(self._labels)
    kv = dict(self)
    txt = f'(:{label} {kv})'
    return txt


def node_label_txt(self):
    vs = list(self._labels)
    return ','.join(vs)


if py2neo:
    py2neo.Node.__str__ = node_str_fn  # 修复Node显示中文乱码的问题
    py2neo.Node.label = node_label_txt  # 添加一个函数方便提取node的label


class NeoTool:
    def __init__(self, host='bolt://localhost:7687', user='neo4j', password='neo'):
        self.graph = py2neo.Graph(host, auth=(user, password))
        logs.brief.info(f'已连结neo4j {user}@{host}')

    def graph_run(self, query):
        resp = self.graph.run(query)
        xx = resp.to_subgraph()
        yy = resp.data()

    def get_node(self, name):
        query = f'MATCH (n) WHERE n.name = "{name}" RETURN n'
        node = self.graph.run(query).to_subgraph()
        return node

    def match_nodes(self, query='', k=None, v=None):
        # 获取 节点
        if k and v:
            query = f' WHERE n.{k} = "{v}"'
        txt = f'MATCH (n){query} RETURN n'
        g = self.graph.run(txt).to_subgraph()
        out = [node for node in g.nodes]
        return out

    def match_rels(self, nodes=None, r_type=None, limit=None):
        # 获取 关系 （图数据库优化的是图计算，也就是节点的连接关系，连通性之类的这些，查询关系比较慢）
        rels = self.graph.match(nodes=nodes, r_type=r_type, limit=limit)
        print(len(rels))
        out = []
        for rel in tqdm(rels, desc='iter relation', leave=False):
            out.append(rel)
        return out

    def delete(self, query):
        # query = 'MATCH (n) WHERE n.name = "张三" RETURN n'
        nodes = self.graph.run(query).to_subgraph()
        self.graph.delete(nodes)

    def add_node(self, label, **kv):
        node = py2neo.Node(label, **kv)
        self.graph.create(node)
        return node

    def add_relation(self, node1, label, node2):
        rel = py2neo.Relationship(node1, label, node2)
        self.graph.create(rel)
        return rel

    @staticmethod
    def query_all_node():
        return 'MATCH (n) RETURN n'

    def add_rel_by_name(self, name1, rel, name2):
        # 通过节点name添加关系
        node1 = self.get_node(name1)
        node2 = self.get_node(name2)
        if node1 is not None and node2 is not None:
            self.add_relation(node1, rel, node2)
            logs.tmp.info(f'{name1}-{rel}->{name2}')
        else:
            logs.tmp.error(f'node1={node1}  node2={node2}')
            raise InfoError('添加关系失败  节点未找到')


if __name__ == '__main__':
    # db = NeoTool()
    # node = db.get_node('name', '吸血鬼')
    # print(node)

    node = py2neo.Node('标签', 权重='中文', 维度=['中文', '中午'])
    print(node)
    # print(str(node))
    # print(node_str(node))



import os
from fish_tool.log_tool import logs, InfoError
import requests.exceptions as e
import urllib3.exceptions
import socket
import uvicorn
# 便于其他代码导入（不能删除）
from pydantic import BaseModel
from fastapi import FastAPI, Response, APIRouter, Header, Request, HTTPException, Depends, Security, Cookie

# 常用的requests异常
net_excepts = (
    e.ReadTimeout, e.ChunkedEncodingError, e.ConnectionError,
    socket.gaierror,
    urllib3.exceptions.NewConnectionError, urllib3.exceptions.MaxRetryError
)

# 设置 uvicorn的日志格式
if uvicorn:
    fmt = '[%(levelname)s %(asctime)s %(filename)s %(funcName)s:%(lineno)d] %(message)s'
    log_config = uvicorn.config.LOGGING_CONFIG
    log_config["formatters"]["default"]["fmt"] = fmt

app_cache = {}


def get_app(app_name='default'):
    from fastapi.middleware.cors import CORSMiddleware
    from fastapi.middleware.gzip import GZipMiddleware
    if app_name not in app_cache:
        # 参数关闭/docs /docs/oauth2-redirect /redoc /openapi.json
        app = FastAPI(docs_url=None, redoc_url=None, openapi_url=None)
        # 允许跨域 便与测试
        app.add_middleware(
            CORSMiddleware,
            allow_origins=['*'],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        # 添加gzip压缩中间件
        app.add_middleware(GZipMiddleware, minimum_size=1000)
        app_cache[app_name] = app
    return app_cache[app_name]


def show_router(app):
    for _router in app.router.routes:
        logs.tmp.info(_router.path or _router.path_format)


def deploy(app, port=80):
    uvicorn.run(app, host='0.0.0.0', port=port)


def add_file_url(app, url, path):
    # 添加静态文件的url
    from fastapi.staticfiles import StaticFiles
    if not os.path.isfile(path):
        raise InfoError(f'静态url 文件不存在={path}')
    scope = {'method': 'GET', 'headers': {}}
    folder = os.path.dirname(path)
    name = os.path.basename(path)
    static = StaticFiles(directory=folder)

    @app.get(url)
    async def aifish_manage():
        return await static.get_response(name, scope)

    # 示范： add_file_url(app, '/favicon.ico', 'data/img/icon.jpg')


def add_dir_url(app, url, path):
    # 添加静态文件夹的url
    from fastapi.staticfiles import StaticFiles
    if not os.path.isdir(path):
        raise InfoError(f'静态url 文件夹不存在={path}')
    app.mount(url, StaticFiles(directory=path), name=url)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import LayerNorm
import math
import copy
from fish_tool.ai import torch_tool


def make_decoder(head_num, d_model, attn_layers, attn_dropout=0.1):
    attn = MultiHeadedAttention(head_num, d_model)
    ff = PositionwiseFeedForward(d_model, d_model, attn_dropout)
    decoder = Decoder(DecoderLayer(d_model, attn, ff, attn_dropout), attn_layers)
    return decoder


def clones(module, N):
    "Produce N identical layers."
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


class Decoder(nn.Module):
    "Generic N layer decoder with masking."

    def __init__(self, layer, N):
        super().__init__()
        self.layers = clones(layer, N)
        self.norm = LayerNorm(layer.d_model)

    def forward(self, x, memory, src_mask=None):
        for layer in self.layers:
            x = layer(x, memory, src_mask)
        return self.norm(x)


class SublayerConnection(nn.Module):
    """
    A residual connection followed by a layer norm.
    Note for code simplicity the norm is first as opposed to last.
    """

    def __init__(self, d_model, dropout):
        super().__init__()
        self.norm = LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, sublayer):
        "Apply residual connection to any sublayer with the same size."
        return x + self.dropout(sublayer(self.norm(x)))


class DecoderLayer(nn.Module):
    "Decoder is made of self-attn, src-attn, and feed forward (defined below)"

    def __init__(self, d_model, src_attn, feed_forward, dropout):
        super().__init__()
        self.d_model = d_model
        self.src_attn = src_attn
        self.feed_forward = feed_forward
        self.sublayer = clones(SublayerConnection(d_model, dropout), 2)

    def forward(self, x, memory, src_mask):
        "Follow Figure 1 (right) for connections."
        m = memory
        x = self.sublayer[0](x, lambda x: self.src_attn(x, m, m, src_mask))
        return self.sublayer[1](x, self.feed_forward)


class MultiHeadedAttention(nn.Module):
    def __init__(self, h, d_model, dropout=0.1):
        "Take in model size and number of heads."
        super().__init__()
        assert d_model % h == 0, f'd_model={d_model}  heads={h}'
        # We assume d_v always equals d_k
        self.d_k = d_model // h
        self.h = h
        self.linears = clones(nn.Linear(d_model, d_model), 4)
        self.attn = None
        self.dropout = nn.Dropout(p=dropout) if dropout else None

    def forward(self, query, key, value, mask=None):
        """
        :param query: (bsz, q_len, dim)
        :param key: (bsz, seq_len, dim)
        :param value: (bsz, seq_len, dim)
        :param mask: (bsz, q_len, seq_len)   等于0的位置会被遮罩
        :return:
        """
        if mask is not None:
            # Same mask applied to all h heads.
            mask = mask.unsqueeze(1)
        nbatches = query.size(0)

        query, key, value = [l(x).view(nbatches, -1, self.h, self.d_k).transpose(1, 2)
                             for l, x in zip(self.linears, (query, key, value))]
        # query ~ (bsz, heads, q_len, head_dim)
        # key ~ (bsz, heads, seq_len, head_dim)
        # value ~ (bsz, heads, seq_len, head_dim)

        # 2) Apply attention on all the projected vectors in batch.
        x, self.attn = attention(query, key, value, mask=mask, dropout=self.dropout)  # x~(bsz, heads, q_len, head_dim)

        # 3) "Concat" using a view and apply a final linear.
        x = x.transpose(1, 2).contiguous()  # x~(bsz, q_len, heads, head_dim)
        x = x.view(nbatches, -1, self.h * self.d_k)  # x~(bsz, q_len, dim)
        return self.linears[-1](x)


class PositionwiseFeedForward(nn.Module):
    "Implements FFN equation."

    def __init__(self, d_model, d_ff, dropout=0.1):
        super().__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        """
        MLP: 输入输出格式不变
        """
        x = self.dropout(F.relu(self.w_1(x)))
        x = self.w_2(x)
        return x


def attention(query, key, value, mask=None, dropout=None):
    """
    Compute 'Scaled Dot Product Attention
    :param query: (bsz, heads, q_len, dim)
    :param key: (bsz, heads, seq_len, dim)
    :param value: (bsz, heads, seq_len, dim)
    :param mask: (bsz, 1, 1, seq_len) or (bsz, 1, q_len, seq_len)  等于0的位置会被遮罩
    :param dropout:
    :return: value~(bsz, heads, q_len, dim), p_attn~(bsz, heads, q_len, seq_len)
    """
    d_k = query.size(-1)
    key = key.transpose(-2, -1)  # key~(bsz,heads,dim,seq_len)
    scores = torch.matmul(query, key) / math.sqrt(d_k)  # scores~(bsz,heads,q_len,seq_len)
    if mask is not None:
        scores = scores.masked_fill(mask == 0, -1e9)  # mask == 0 的位置才mask为很小的值
    p_attn = F.softmax(scores, dim=-1)  # p_attn~(bsz,heads,q_len,seq_len)
    if dropout is not None:
        p_attn = dropout(p_attn)
    value = torch.matmul(p_attn, value)  # value~(bsz,heads,q_len,dim)
    return value, p_attn


class TransformerDecoder(nn.Module):
    def __init__(self, N, d_model, d_ff, h, dropout):
        super().__init__()
        attn = MultiHeadedAttention(h, d_model)
        ff = PositionwiseFeedForward(d_model, d_ff, dropout)
        self.attention_layer = Decoder(DecoderLayer(d_model, attn, ff, dropout), N)

    def forward(self, input_q, input_k, att_m):
        attention_output = self.attention_layer(input_q, input_k, att_m)
        # (batch, num_query, hidden)
        return attention_output


def test_decoder():
    head_num = 2
    dim = 10
    dropout = 0.1
    decoder_layers = 2
    decoder = make_decoder(head_num, dim, decoder_layers, attn_dropout=dropout)
    bsz = 1
    seq_len = 6
    q_len = 4
    query = torch.rand(bsz, q_len, dim)
    value = torch.rand(bsz, seq_len, dim)
    mask = torch.tensor([[1, 1, 1, 1, 0, 0]], dtype=torch.long)
    y = decoder(query, value, mask)  # ~ (bsz, q_len, dim)
    print(y)


class AttentionPool(torch.nn.Module):
    # 把序列向量 汇聚成 N维logit输出
    def __init__(self, out_dim, dim=768, n_head=2, dropout=0.1):
        super().__init__()
        self.query = torch.nn.Parameter(torch.rand(1, out_dim, dim, requires_grad=True), requires_grad=True)
        self.attn = MultiHeadedAttention(h=n_head, d_model=dim)
        self.line = torch.nn.Linear(dim, 1)
        self.dropout = torch.nn.Dropout(p=dropout) if dropout else None
        torch_tool.init_linear(self.line)

    def forward(self, x):
        # x~(bsz, seq_len, dim)   logit~(bsz, out_dim)
        bsz, seq_len, dim = x.size()
        query = self.query.expand(bsz, -1, -1)  # (bsz, out_dim, dim)
        pool_x = self.attn(query, x, x)  # pool_x~(bsz, out_dim, dim)
        if self.dropout is not None:
            pool_x = self.dropout(pool_x)
        logit = self.line(pool_x)  # ~(bsz, out_dim, 1)
        logit = logit.squeeze(2)  # ~(bsz, out_dim)
        return logit


if __name__ == '__main__':
    test_decoder()



import json
import zipfile
from collections import Counter


def read_xmind(path):
    f = zipfile.ZipFile(path, 'r')
    data = json.loads(f.read('content.json'))
    return data


def get_leafs(data, pre=''):
    out = []
    if isinstance(data, list):
        for one in data:
            out.extend(get_leafs(one, pre))
        return out
    if 'rootTopic' in data:
        data = data['rootTopic']
    title = data['title']
    if 'children' in data and 'attached' in data['children']:
        pre = f'{pre}_{title}' if pre else title
        for sub in data['children']['attached']:
            out.extend(get_leafs(sub, pre))
    else:
        title = data['title']
        info = {
            'id': data['id'],
            'title': title,
            'join_title': f'{pre}_{title}' if pre else title
        }
        if 'labels' in data:
            info['labels'] = data['labels']
        if 'children' in data and 'callout' in data['children']:
            info['callout'] = data['children']['callout']
        out.append(info)
    return out


def check_leaf_is_same(path):
    # 检查标签体系的叶子节点的文本是否有重复的
    data = read_xmind(path)
    leafs = get_leafs(data)
    leaf__count = Counter([t['title'] for t in leafs])
    print(f'leaf__count={leaf__count}')
    print('=' * 60)
    for leaf, c in leaf__count.items():
        if c > 1:
            print('EORROR 重复的标签： ', leaf, c)


if __name__ == '__main__':
    check_leaf_is_same(path='C:/Users/fish/Desktop/标签体系/小说标签体系8.xmind')



# 为了方便引用 载入常用对象
from fish_tool.log_tool import logs, InfoError
from fish_tool.config_base import BaseConfig

"""   依赖关系
log_tool
    无
    
sys_tool
    log_tool
register
    log_tool
config_base
    log_tool
    
ai.laboratory
    log_tool
ai.torch_tool
    log_tool
    
db.es_tool
    log_tool
db.sqlite_tool
    log_tool
db.neo_tool
    log_tool
    
net.spider_tool
    log_tool
net.proxy_16yun.py
    log_tool
    
"""



import inspect
from pprint import pformat
import logging.handlers
import os
from os.path import join
import tqdm

try:
    from fastapi import HTTPException
except ImportError:
    HTTPException = Exception

logging.basicConfig(level=logging.INFO, format='[%(levelname)s %(filename)s %(funcName)s:%(lineno)d] %(message)s')

long_fmt = logging.Formatter(fmt='[%(asctime)s %(levelname)s %(filename)s %(funcName)s:%(lineno)d] %(message)s', datefmt='%H:%M:%S')
long_time_fmt = logging.Formatter(fmt='[%(asctime)s %(levelname)s %(filename)s %(funcName)s:%(lineno)d] %(message)s', datefmt=None)


class InfoError(HTTPException):
    def __init__(self, detail, status_code=500):
        if HTTPException is not None:
            super().__init__(status_code=status_code, detail=detail)
        else:
            super().__init__(detail)


class TqdmLoggingHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


# 标准输出（正常也就是打印到屏幕, 不会破坏tqdm进度条）
std_handler = TqdmLoggingHandler(logging.DEBUG)
std_handler.setFormatter(long_fmt)


def remove_handler(logger):
    for h in list(logger.handlers):
        logger.removeHandler(h)


class Logs:
    # 临时日志（只在屏幕上显示，不存储到文件）
    tmp = logging.getLogger('log_tmp')
    tmp.setLevel(logging.DEBUG)
    tmp.addHandler(std_handler)
    tmp.propagate = False

    def __init__(self, folder, level='INFO'):
        # brief 记录简要日志， 通常是统计分数等 (也会存储到all的日志文件里，时all的上下文更加完整）
        # all 记录细节日志， 全部存储到文件里，通常是训练过程中的细节，便于发现错误和debug
        # rotate 滚动细节日志， 滚动存储到文件里， 通常是部署服务后的日志，因为持续时间长，所以要滚动存储，避免占满硬盘
        self.folder = os.path.abspath(folder)
        os.makedirs(self.folder, exist_ok=True)
        self.brief = logging.getLogger('log_brief')
        self.all = logging.getLogger('log_all')
        self.rotate = logging.getLogger('log_rotate')

        self.set_logger(folder, level)

    def print(self, *args, width=200, show_path=-3):
        caller = inspect.stack()[1]
        paths = caller.filename.split('/')
        file = '/'.join(paths[show_path:])
        pre = f'[{file} {caller.function}:{caller.lineno}]'
        for arg in args:
            print(f'{pre}{pformat(arg, width=width)}')

    def reset(self, folder, level='INFO'):
        # 重置日志存储位置
        folder = os.path.abspath(folder)
        if self.folder != folder:
            self.folder = folder
            os.makedirs(folder, exist_ok=True)
            remove_handler(self.brief)
            remove_handler(self.all)
            remove_handler(self.rotate)
            self.set_logger(folder, level)

    def set_logger(self, folder, level='INFO'):
        handler_brief = logging.FileHandler(join(folder, 'log_brief.log'), mode='a', encoding='utf8', delay=True)
        handler_brief.setLevel(level)
        handler_brief.setFormatter(long_fmt)

        handler_all = logging.FileHandler(join(folder, 'log_detail_all.log'), mode='a', encoding='utf8', delay=True)
        handler_all.setLevel(level)
        handler_all.setFormatter(long_time_fmt)

        handler_rotate = logging.handlers.RotatingFileHandler(join(folder, 'log_detail_rotate.log'),
                                                              maxBytes=100000, backupCount=10, mode='a',
                                                              encoding='utf8', delay=True)
        handler_rotate.setLevel(level)
        handler_rotate.setFormatter(long_time_fmt)

        self.brief.setLevel(level)
        self.brief.addHandler(std_handler)
        self.brief.addHandler(handler_brief)
        self.brief.addHandler(handler_all)
        self.brief.propagate = False

        self.all.setLevel(level)
        self.all.addHandler(std_handler)
        self.all.addHandler(handler_all)
        self.all.propagate = False

        self.rotate.setLevel(level)
        self.rotate.addHandler(std_handler)
        self.rotate.addHandler(handler_rotate)
        self.rotate.propagate = False


tmp_folder = join(os.path.dirname(__file__), '../logs')
logs = Logs(tmp_folder)  # 初始化一个临时位置 存储日志

# 可以在其他位置使用 logs.reset(folder)  重置日志存储位置



import random, requests, httpx
import logging
from fish_tool.log_tool import logs, InfoError


# 代理验证信息
web_info = '115.231.107.229:6460:16WASOTM:085764'
web_infos = web_info.split(':')
proxyUser = web_infos[2]  # "16WASOTM"
proxyPass = web_infos[3]  # "085764"
proxyHost = f'{web_infos[0]}:{web_infos[1]}'  # "115.231.107.229:6460"

proxyMeta = f'http://{proxyUser}:{proxyPass}@{proxyHost}'

ua_list = [
    'Mozilla/5.0 (compatible; MSIE 10.0; qdesk 2.4.1266.203; Windows NT 6.1; WOW64; Trident/6.0)',
    'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.2; Win64; x64; Trident/6.0; Touch; MAARJS)',
    'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.2; WOW64; Trident/6.0; Touch; QQBrowser/8.0.3345.400)',
    'Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0; .NET CLR 1.1.4322; .NET CLR 2.0.50727)',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10) AppleWebKit/600.1.25 (KHTML, like Gecko) Version/8.0 Safari/600.1.25',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_4) AppleWebKit/601.5.17 (KHTML, like Gecko) Version/9.1 Safari/601.5.17',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/601.6.17 (KHTML, like Gecko) Version/9.1.1 Safari/601.6.17',
    'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/72.0.3626.109 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_5_8) AppleWebKit/534.24 (KHTML, like Gecko) Chrome/11.0.696.68 Safari/534.24',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_5_8) AppleWebKit/534.31 (KHTML, like Gecko) Chrome/13.0.748.0 Safari/534.31',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_5_8) AppleWebKit/535.1 (KHTML, like Gecko) Chrome/14.0.801.0 Safari/535.1',
]


def get(url, proxy=False):
    ua = random.choice(ua_list)
    headers = {'User-Agent': ua}
    if proxy:
        tunnel = random.randint(1, 900000)
        headers["Proxy-Tunnel"] = str(tunnel)
        proxies = {"http": proxyMeta, "https": proxyMeta}
    else:
        proxies = None
    resp = requests.get(url, proxies=proxies, headers=headers, timeout=3)
    return resp


async def async_get(url, proxy=False):
    tunnel = random.randint(1, 800000)
    ua = random.choice(ua_list)
    headers = {"Proxy-Tunnel": str(tunnel), 'User-Agent': ua}
    if proxy:
        proxies = {"http": proxyMeta, "https": proxyMeta}
    else:
        proxies = None
    async with httpx.AsyncClient(proxies=proxies, headers=headers, timeout=3) as client:
        # resp = await client.get(url, headers=headers, timeout=3)
        resp = await client.get(url)
    return resp



import inspect
from pprint import pprint, pformat
from copy import deepcopy
import math
import time
import sys
import traceback
import os
import json
import pickle
import requests
from fish_tool.db.zip_tool import ZipTool
from tqdm import tqdm
import httpx
from fish_tool.log_tool import logs, InfoError


class ScrollIter:
    # 包装一个迭代器， 使其增加长度属性， 便于tqdm等显示进度条
    def __init__(self, host, index, query, wait='3m', timeout=20000):
        self.return_num = 0
        self.wait = wait
        url = f'{host}/{index}/_doc/_search?scroll={wait}'
        resp = requests.post(url, json=query, timeout=timeout).json()

        self.url_scroll = f'{host}/_search/scroll'
        try:
            self.total = resp['hits']['total']['value']
            self._scroll_id = resp['_scroll_id']
            self.hits = resp['hits']['hits']
        except:
            logs.rotate.error(f'url={url}')
            logs.rotate.error(f'query={query}')
            logs.rotate.error(f'resp={resp}')
            raise

    def __len__(self):
        return self.total

    def __iter__(self):
        while self.return_num < self.total:
            if self.hits:
                for one in self.hits:
                    self.return_num += 1
                    yield one
            body_scroll = {'scroll': self.wait, 'scroll_id': self._scroll_id}
            resp = requests.post(self.url_scroll, json=body_scroll).json()
            self._scroll_id = resp['_scroll_id']
            self.hits = resp['hits']['hits']


class ScrollBatch:
    def __init__(self, es, index, query, wait='3m', timeout=20000):
        self.es = es
        self.wait = wait
        url = f'{self.es.host}/{index}/_doc/_search?scroll={wait}'
        self.url_scroll = '{}/_search/scroll'.format(self.es.host)
        resp = requests.post(url, json=query, timeout=timeout).json()
        try:
            self.total = math.ceil(resp['hits']['total']['value'] / query.get('size', 10))
            self._scroll_id = resp['_scroll_id']
            self.hits = resp['hits']['hits']
        except:
            logs.rotate.error(f'url={url}')
            logs.rotate.error(f'query={query}')
            logs.rotate.error(f'resp={resp}')
            raise

    def __len__(self):
        return self.total

    def __iter__(self):
        while self.hits:
            yield self.hits
            body_scroll = {'scroll': self.wait, 'scroll_id': self._scroll_id}
            resp = requests.post(self.url_scroll, json=body_scroll).json()
            self._scroll_id = resp['_scroll_id']
            self.hits = resp['hits']['hits']


class Query:
    def __init__(self, _query):
        self._query = self.extract_query(_query)
        self.functions = []

    def __call__(self, size=20, start=0, sort='', desc=True, _source=None, excludes=None, fn_boost_mode='multiply', fn_boost=None, fn_min_score=None):
        # desc是降序排序（sort有参数时才有用）
        # _source 返回数据的字段数组
        # fn_boost 函数分数的boost   fn_boost_mode 函数和query的分数结合模式   fn_min_score最小值（会筛选掉不符合的doc）
        if self.functions:
            assert fn_boost_mode in ['multiply', 'replace', 'sum', 'avg', 'max', 'min']
            if len(self.functions) == 1:
                fn = deepcopy(self.functions)[0]
                fn['boost'] = fn.pop('weight', 1)  # 一个函数则字段是boost  多个函数则字段是weight
                fn['query'] = self._query
                fn['boost_mode'] = fn_boost_mode
                _query = {'function_score': fn}
            else:
                _query = {
                    'function_score': {
                        'query': self._query,
                        'functions': self.functions,
                        'boost_mode': fn_boost_mode,
                    }
                }
            if fn_boost is not None:
                _query['function_score']['boost'] = fn_boost
            if fn_min_score is not None:
                _query['function_score']['min_score'] = fn_min_score
            out = {'query': _query}
        else:
            out = {'query': self._query}
        out['size'] = size
        out['from'] = start
        if sort:
            order = {'order': 'desc' if desc else 'asc'}
            out['sort'] = [{sort: order}]
        body_source = {}
        if _source:
            if isinstance(_source, str):
                body_source["includes"] = [_source]
            elif isinstance(_source, list):
                body_source["includes"] = _source
            if isinstance(excludes, str):
                body_source["excludes"] = [excludes]
            elif isinstance(excludes, list):
                body_source["excludes"] = excludes
        if body_source:
            out['_source'] = body_source
        return out

    @classmethod
    def _recurrence_json_to_query(cls, kv):
        if isinstance(kv, dict):
            if '_fn' in kv:
                fn = getattr(Query, kv['_fn'])
                param = Query._recurrence_json_to_query(kv.get('_param', {}))
                query = fn(**param)
                if '_call' in kv:
                    query = query(**kv['_call'])
                return query
            else:
                return {k: Query._recurrence_json_to_query(v) for k, v in kv.items()}
        if isinstance(kv, (list, tuple, set)):
            return [Query._recurrence_json_to_query(t) for t in kv]
        else:
            return kv

    @classmethod
    def json_to_query(cls, kv):
        # 通过json字典转换成query（便于不能直接写代码的环境使用， 例如客户端调用）
        # {‘_fn’: 'query函数名', '_call': {query(参数字典)}， ‘_param’： {_fn（参数字典）}}
        kv = deepcopy(kv)
        query = cls._recurrence_json_to_query(kv)
        if isinstance(query, Query):
            query = query()
        return query

    @classmethod
    def all(cls, size=20, start=0, sort='', desc=True, _source=None, _name='', excludes=None):
        # 直接返回的是query的字典
        q = {"match_all": {}}
        if _name:
            q['match_all']['_name'] = _name
        query = cls(q)
        return query(size=size, start=start, sort=sort, desc=desc, _source=_source, excludes=excludes)

    @staticmethod
    def extract_query(query):
        # 提取内部的query字典
        if isinstance(query, Query):
            query = query._query
        if 'query' in query:
            query = query['query']
        return query

    @classmethod
    def analyze(cls):
        url = 'http://{host}:{port}/{index}/_analyze'.format(host='', port='', index='')
        return url

    def fn_decay(self, fn, key, weight=1, origin=None, offset=None, scale=None, decay=0.5):
        # fn=空文本则函数的固定分数=boost
        assert fn in ['gauss', 'exp', 'linear']
        one = {'weight': weight}
        if fn:
            one[fn] = {
                key: {
                    "origin": origin,
                    "scale": scale,
                    "offset": offset,
                    "decay": decay
                }
            }
        self.functions.append(one)

    @classmethod
    def aggs(cls, name__field, query=None, size=100):
        # field文本的末尾记得添加 .keyword
        # 直接返回query数据 而不是Query对象
        # name__field = {name: field_str}                       (类型默认是terms)
        # name__field = {name: {field: field_str, size: int}}   (类型默认是terms)
        # name__field = {name: {field: field_str, size: int, type: 'terms'}}
        # name__field = {name: {'type': 'histogram', 'field': '字段', 'interval': 1}}  数值按照间隔=1进行统计
        # name__field = {name: {'type': 'cardinality', 'field': '字段'}}  统计 桶的数量
        q = {
            'size': 0,
            'aggs': {
                # name: {
                #     'terms': {
                #         'field': field,
                #         'size': 20,
                #     },
                # 'aggs': 子聚合
                # },
            }
        }
        if query:
            if isinstance(query, Query):
                query = query._query
            if isinstance(query, dict):
                if 'query' in query:
                    query = query['query']
            q['query'] = query
        for name, field in name__field.items():
            if isinstance(field, str):
                wrap = {'terms': {'field': field, 'size': size}}
            else:
                agg_type = field.pop('type', 'terms')
                wrap = {agg_type: field}

                if 'aggs' in field:
                    wrap['aggs'] = field.pop('aggs')
                    wrap['size'] = size
            q['aggs'][name] = wrap
        return q

    @classmethod
    def term(cls, key, values, boost=None, boost_change=True, _name=''):
        if isinstance(values, (int, float)):
            q = {"term": {
                key: {
                    "value": values,
                }}}

            if boost is not None or boost_change:
                if boost is None:
                    boost = 1.0
                q['term'][key]['boost'] = boost
        elif isinstance(values, str):
            q = {"term": {
                key: {
                    "value": values,
                }}}
            if boost is not None or boost_change:
                if boost is None:
                    boost = 1.0
                if boost_change:
                    total = len(values)
                    if total:
                        boost /= len(values)  # 根据搜索文本长度归一化
                q['term'][key]['boost'] = boost
        elif len(values) == 1:
            q = {"term": {
                key: {
                    "value": values[0],
                    "boost": boost
                }}}
            if boost is not None or boost_change:
                if boost is None:
                    boost = 1.0
                if boost_change:
                    total = len(values[0])
                    if total:
                        boost /= len(values[0])  # 根据搜索文本长度归一化
                q['term'][key]['boost'] = boost
        else:
            should = [cls.term(key, v, boost_change=boost_change) for v in values]
            if boost is not None or boost_change:
                if boost is None:
                    boost = 1.0
                if boost_change:
                    total = len(values)
                    if total:
                        boost /= len(values)  # 根据搜索文本数量归一化
            q = cls.bool(should=should, boost=boost)
        if _name:
            q['term'][key]['_name'] = _name
        query = Query(q)
        return query

    @classmethod
    def term_filter(cls, key, values, required_matches=None, _name=''):
        # 问题： terms是筛选方法，所有分数都等于1   匹配的项数没有区别
        if isinstance(values, list):
            if len(values) == 1:
                q = {"term": {key: values[0]}}
            else:
                q = {"terms": {key: list(values)}}
        else:
            q = {"term": {key: values}}
        if required_matches is not None:
            q['required_matches'] = required_matches
        if _name:
            if 'term' in q:
                q['term'][key]['_name'] = _name
            else:
                q['terms'][key]['_name'] = _name
        query = Query(q)
        return query

    @classmethod
    def match(cls, key, value, boost=None, boost_change=True, analyzer=None, _name=''):
        q = {"match": {key: {"query": value}}}
        if _name:
            q['match'][key]['_name'] = _name
        if analyzer:
            q['match'][key]['analyzer'] = analyzer
        if boost is not None or boost_change:
            if boost is None:
                boost = 1.0
            if boost_change:
                total = len(value)
                if total:
                    boost /= len(value)  # 根据搜索文本长度归一化
            q['match'][key]['boost'] = boost
        query = Query(q)
        return query

    @classmethod
    def match_phrase(cls, key, value, boost=None, _name=''):
        q = {"match_phrase": {key: {"query": value}}}
        if _name:
            q['match_phrase'][key]['_name'] = _name
        if boost is not None:
            q['match_phrase'][key]['boost'] = boost
        return Query(q)

    @classmethod
    def range(cls, key, start=None, end=None, boost=1.0, _name=''):
        # 示范： cls.range('a', '>3', '<=10')
        q = {
            "range": {
                key: {
                    "boost": boost
                }
            }
        }
        if _name:
            q['range'][key]['_name'] = _name
        if start is not None:
            q['range'][key]['gte'] = start
        if end is not None:
            q['range'][key]['lte'] = end
        query = Query(q)
        return query

    @classmethod
    def nest(cls, path, query):
        q = {
            "nested": {
                "path": path,
                "query": Query.extract_query(query)
            }
        }
        query = Query(q)
        return query

    @classmethod
    def exist(cls, key, _name=''):
        q = {
            "exists": {
                "field": key
            }
        }
        if _name:
            q['exists']['_name'] = _name
        query = Query(q)
        return query

    @classmethod
    def geo_filter(cls, key, location, distance, boost=None):
        # 按地理距离筛选, 符合的分数=1 否则分数=0
        q = {
            "geo_distance": {
                "distance": distance,  # 10000m   = 10000米
                key: location,  # {"lat": 39.504901,"lon": 118.966,},
            }
        }
        if boost:
            q['geo_distance']['boost'] = boost
        query = Query(q)
        return query

    @classmethod
    def bool(cls,
             must: list = None,
             should: list = None,
             must_not: list = None,
             filter: list = None,
             boost: float = None,
             boost_change: bool = True,
             minimum_should_match: int = 0,
             _name: str = '',
             ):
        q = {"bool": {}}
        lengths = []
        if must:
            lengths.append(len(must))
            q['bool']['must'] = [Query.extract_query(t) for t in must]
        if should:
            lengths.append(len(should))
            q['bool']['should'] = [Query.extract_query(t) for t in should]
        if must_not:
            q['bool']['must_not'] = [Query.extract_query(t) for t in must_not]
        if filter:
            q['bool']['filter'] = [Query.extract_query(t) for t in filter]
        if boost is not None or boost_change:
            if boost is None:
                boost = 1.0
            if boost_change:
                total = sum(lengths)
                if total:
                    boost /= total  # 根据查询子句的总数量归一化
            q['bool']['boost'] = boost
        if minimum_should_match:
            q['bool']['minimum_should_match'] = minimum_should_match
        if _name:
            q['bool']['_name'] = _name
        query = BoolQuery(q)
        return query

    @classmethod
    def boost(cls, positive: dict, negative: dict, negative_boost: float = 0.2):
        q = {
            "boosting": {
                "positive": positive,
                "0": negative,
                "negative_boost": negative_boost
            }
        }
        query = Query(q)
        return query


class BoolQuery(Query):
    def __init__(self, _query):
        super().__init__(_query)

    def __call__(self, *args, **kv):
        if any(self._query['bool'].values()):
            fn = super().__call__
        else:
            fn = Query.all.__call__
            # 没有任何条件 则query相当于 all
        return fn(*args, **kv)

    def must(self, query):
        self._query['bool'].setdefault('must', []).append(Query.extract_query(query))

    def should(self, query):
        self._query['bool'].setdefault('should', []).append(Query.extract_query(query))

    def no(self, query):
        self._query['bool'].setdefault('must_not', []).append(Query.extract_query(query))

    def filter(self, query):
        self._query['bool'].setdefault('filter', []).append(Query.extract_query(query))


class EsTool:
    def __init__(self, host):
        self.host = host
        self.headers = {'Content-Type': 'application/json'}

    def all_index(self):
        url = f'{self.host}/_cat/indices?format=json'
        resp = requests.get(url, timeout=12)
        out = resp.json()
        return out

    def explain(self, index, query, doc_id=None):
        if doc_id:
            url = f'{self.host}/{index}/_explain/{doc_id}'
            query = {k: v for k, v in query.items() if k in ['query']}  # 去除 size 等无用的参数
        else:
            url = f'{self.host}/{index}/_doc/_search'
            query = deepcopy(query)
            query['explain'] = True
        resp = requests.post(url, json=query, headers=self.headers, timeout=3)
        resp_obj = resp.json()
        out = resp_obj
        return out

    def waiting_es_is_ok(self, minute=10):
        for i in range(int(minute * 60)):
            try:
                resp = requests.get(self.host, timeout=1)
                txt = resp.text
            except:
                txt = '404'
            if 'You Know, for Search' in txt:
                logs.rotate.warning(f'ES is OK!')
                return
            else:
                logs.rotate.warning(f'ES is not OK. It is Waiting!')
                time.sleep(1)
        raise ValueError(f'{self.host} 无法连结到ES')

    async def async_count(self, index, query=None):
        url = f'{self.host}/{index}/_doc/_count'
        if query is None:
            query = Query.all()
        elif isinstance(query, Query):
            query = query()
        query = {k: v for k, v in query.items() if k in ['query']}
        try:
            async with httpx.AsyncClient() as client:
                resp = await client.post(url, json=query, headers=self.headers, timeout=3)
            resp = resp.json()
            if 'error' in resp:
                logs.rotate.error(f'error={resp}')
            out = resp.get('count', 0)
            return out
        except Exception:
            raise

    def count(self, index, query=None):
        url = f'{self.host}/{index}/_doc/_count'
        if query is None:
            query = Query.all()
        elif isinstance(query, Query):
            query = query()
        query = {k: v for k, v in query.items() if k in ['query']}
        try:
            resp = requests.post(url, json=query, headers=self.headers, timeout=3)
            resp = resp.json()
            if 'error' in resp:
                logs.rotate.error(f'error={resp}')
            out = resp.get('count', 0)
            return out
        except Exception:
            raise

    async def async_save(self, doc, index, doc_id=''):
        # 增 （如果id已存在则替换内容） 如果id和doc都存在则update
        if doc_id:
            url = f'{self.host}/{index}/_doc/{doc_id}/_update'
        else:
            url = f'{self.host}/{index}/_doc/'
        try:
            async with httpx.AsyncClient() as client:
                if doc_id:
                    body = {'doc': doc, 'doc_as_upsert': True}
                    resp = await client.post(url, json=body, headers=self.headers, timeout=3)
                else:
                    resp = await client.post(url, json=doc, headers=self.headers, timeout=3)
            resp_obj = resp.json()
            # 'result': 'created'  创建成功   updated更新成功
            # 'status': 409        已存在了，冲突，创建失败
            logs.rotate.info(f'save_resp={resp_obj}')
            out = resp_obj
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}  body={doc}  超时错误')
            out = {}
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  body={doc}  {error}')
            out = {}
        return out

    def save(self, doc, index, doc_id=''):
        # 警告： 会全文替换， 缺少的字段会被删除
        # 增 （如果id已存在则替换内容）
        if doc_id:
            url = f'{self.host}/{index}/_doc/{doc_id}'
        else:
            url = f'{self.host}/{index}/_doc/'
        try:
            if doc_id:
                resp = requests.put(url, json=doc, headers=self.headers, timeout=3)
            else:
                resp = requests.post(url, json=doc, headers=self.headers, timeout=3)
            resp_obj = resp.json()
            out = resp_obj
            # id重复错误
            # {'error': {'root_cause': [{'type': 'version_conflict_engine_exception', 'reason': '[a001]: version
            # conflict, document already exists (current version [1])', 'index_uuid': 'BL8rKRXgRkKX3LgKb7MPEg',
            # 'shard': '0', 'index': 'test'}], 'type': 'version_conflict_engine_exception', 'reason': '[a001]: version conflict,
            # document already exists (current version [1])', 'index_uuid': 'BL8rKRXgRkKX3LgKb7MPEg', 'shard': '0', 'index': 'test'}, 'status': 409}
            # 创建成功
            # {'_index': 'test', '_type': '_doc', '_id': 'a002', '_version': 1, 'result': 'created', '_shards': {'total': 2,
            # 'successful': 1, 'failed': 0}, '_seq_no': 1, '_primary_term': 1}

            # {'_index': 'test', '_type': '_doc', '_id': 'nJs2znMBLUO2t0Glmcsb', '_version': 1, 'result': 'created',
            # '_shards': {'total': 2, 'successful': 1, 'failed': 0}, '_seq_no': 2, '_primary_term': 1}
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}  body={doc}  超时错误')
            out = {}
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  body={doc}  {error}')
            out = {}
        return out

    async def async_update_doc(self, doc, index, doc_id, action='update', show_log=True, wait=False, timeout=10):
        # 更新成功返回 True 否则返回False
        assert action in ['update', 'index']
        # logs.rotate.info(f'update_doc={doc}')
        if action == 'update':
            url = f'{self.host}/{index}/_doc/{doc_id}/_{action}'
            body = {'doc': doc}
        else:
            url = f'{self.host}/{index}/_doc/{doc_id}'
            body = doc
        if wait:
            url += '?refresh=wait_for'  # 等待ES存储成功
        try:
            if action == 'update':
                async with httpx.AsyncClient() as client:
                    resp = await client.post(url, json=body, headers=self.headers, timeout=timeout)
            else:
                async with httpx.AsyncClient() as client:
                    resp = await client.put(url, json=body, headers=self.headers, timeout=timeout)
            resp_obj = resp.json()
            if show_log:
                code_caller = inspect.stack()[1]
                code_file = '/'.join(code_caller.filename.split('/')[-3:])
                pre = f'[{code_file} {code_caller.function}:{code_caller.lineno}]'
                logs.rotate.info(f'{pre} update_doc resp_obj={resp_obj}')
            out = resp_obj.get('result') in ['updated', 'noop', 'created']  # noop=更新后未变化，所以没有操作
            # resp原始值
            # 更新成功
            # {'_index': 'test', '_type': '_doc', '_id': 'nJs2znMBLUO2t0Glmcsb', '_version': 2, 'result': 'updated',
            # '_shards': {'total': 2, 'successful': 1, 'failed': 0}, '_seq_no': 3, '_primary_term': 1}
            # 不存在
            # {'error': {'root_cause': [{'type': 'document_missing_exception', 'reason': '[_doc][nJs2znMBLUO2t0Glmcsb--]:
            # document missing', 'index_uuid': 'BL8rKRXgRkKX3LgKb7MPEg', 'shard': '0', 'index': 'test'}], 'type': 'document_missing_exception',
            # 'reason': '[_doc][nJs2znMBLUO2t0Glmcsb--]: document missing', 'index_uuid': 'BL8rKRXgRkKX3LgKb7MPEg', 'shard': '0', 'index': 'test'},
            # 'status': 404}
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}  body={body}  超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  body={body}  {error}')
            out = None
        return out

    def update_doc(self, doc, index, doc_id, action='update', show_log=True, wait=False, timeout=10):
        # 更新成功返回 True 否则返回False
        assert action in ['update', 'index']
        # logs.rotate.info(f'update_doc={doc}')
        if action == 'update':
            url = f'{self.host}/{index}/_doc/{doc_id}/_{action}'
            body = {'doc': doc}
        else:
            url = f'{self.host}/{index}/_doc/{doc_id}'
            body = doc
        if wait:
            url += '?refresh=wait_for'  # 等待ES存储成功
        try:
            if action == 'update':
                resp = requests.post(url, json=body, headers=self.headers, timeout=timeout)
            else:
                resp = requests.put(url, json=body, headers=self.headers, timeout=timeout)
            resp_obj = resp.json()
            if show_log:
                code_caller = inspect.stack()[1]
                code_file = '/'.join(code_caller.filename.split('/')[-3:])
                pre = f'[{code_file} {code_caller.function}:{code_caller.lineno}]'
                logs.rotate.info(f'{pre} update_doc resp_obj={resp_obj}')
            out = resp_obj.get('result') in ['updated', 'noop', 'created']  # noop=更新后未变化，所以没有操作
            # resp原始值
            # 更新成功
            # {'_index': 'test', '_type': '_doc', '_id': 'nJs2znMBLUO2t0Glmcsb', '_version': 2, 'result': 'updated',
            # '_shards': {'total': 2, 'successful': 1, 'failed': 0}, '_seq_no': 3, '_primary_term': 1}
            # 不存在
            # {'error': {'root_cause': [{'type': 'document_missing_exception', 'reason': '[_doc][nJs2znMBLUO2t0Glmcsb--]:
            # document missing', 'index_uuid': 'BL8rKRXgRkKX3LgKb7MPEg', 'shard': '0', 'index': 'test'}], 'type': 'document_missing_exception',
            # 'reason': '[_doc][nJs2znMBLUO2t0Glmcsb--]: document missing', 'index_uuid': 'BL8rKRXgRkKX3LgKb7MPEg', 'shard': '0', 'index': 'test'},
            # 'status': 404}
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}  body={body}  超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  body={body}  {error}')
            out = None
        return out

    async def async_read(self, index, doc_id, _source=None):
        """
        读取文件内容返回， 如果不存在则返回值=default
        :param index:
        :param doc_id:
        :param _source:
        :return:
            {'_index': '', '_type': '_doc', '_id': '', '_primary_term': 3, 'found': True, '_source': {}}
            {'_index': '', '_type': '_doc', '_id': '', 'found': False}
        """
        url = f'{self.host}/{index}/_doc/{doc_id}'
        if _source:
            if isinstance(_source, list):
                _source = ','.join(_source)
            url += f'?_source={_source}'
        try:
            async with httpx.AsyncClient() as client:
                resp = await client.get(url, headers=self.headers, timeout=3)
            out = resp.json()
            if not out.get('found'):
                logs.rotate.error(f'url={url}  {out}')
                out = {'found': False, 'error': f'not found'}
        except requests.exceptions.ReadTimeout:
            error = f'url={url}  超时错误'
            logs.rotate.error(error)
            out = {'found': False, 'error': error}
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  {error}')
            out = {'found': False, 'error': error}
        return out

    def read(self, index, doc_id, _source=None):
        """
        读取文件内容返回， 如果不存在则返回值=default
        :param index:
        :param doc_id:
        :param _source:
        :return:
            {'_index': '', '_type': '_doc', '_id': '', '_primary_term': 3, 'found': True, '_source': {}}
            {'_index': '', '_type': '_doc', '_id': '', 'found': False}
        """
        url = f'{self.host}/{index}/_doc/{doc_id}'
        if _source:
            if isinstance(_source, list):
                _source = ','.join(_source)
            url += f'?_source={_source}'
        try:
            resp = requests.get(url, headers=self.headers, timeout=3)
            out = resp.json()
        except requests.exceptions.ReadTimeout:
            error = f'url={url}  超时错误'
            logs.rotate.error(error)
            out = {'found': False, 'error': error}
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  {error}')
            out = {'found': False, 'error': error}
        return out

    async def async_exist(self, index, doc_id):
        url = f'{self.host}/{index}/_doc/{doc_id}'
        try:
            async with httpx.AsyncClient() as client:
                resp = await client.head(url, headers=self.headers, timeout=3)
            out = resp.status_code == 200
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}  超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  {error}')
            out = None
        return out

    def exist(self, index, doc_id):
        url = f'{self.host}/{index}/_doc/{doc_id}'
        try:
            resp = requests.head(url, headers=self.headers, timeout=3)
            out = resp.status_code == 200
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}  超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  {error}')
            out = None
        return out

    def index_exist(self, index):
        url = f'{self.host}/{index}'
        try:
            resp = requests.get(url, headers=self.headers, timeout=3)
            out = 'error' not in resp.json()
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}   超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  {error}')
            out = None
        return out

    def create_index(self, index):
        if self.index_exist(index):
            return True
        else:
            url = f'{self.host}/{index}'
            resp = requests.put(url)
            out = resp.json()
            return out

    def set_disk(self, low=0.999998, high=0.999999, flood_stage=0.999999):
        # 避免硬盘使用率超过95%时自动进入只读模式
        assert flood_stage >= high
        body = {
            "persistent": {  # persistent表示为永久修改，重启以后也会保存设置   transient表示临时修改，重启以后不会保存设置
                # 这两个是磁盘使用率限制，当磁盘使用率大于低的限制时，如果没有别的node可以存储数据，状态就会变为red
                "cluster.routing.allocation.disk.watermark.low": f"{(low * 100):0.6f}%",  # 默认值 90%
                "cluster.routing.allocation.disk.watermark.high": f"{(high * 100):0.6f}%",  # 默认值 95%

                # 硬盘使用率大于该值， ES会进入只读模式
                "cluster.routing.allocation.disk.watermark.flood_stage": f"{(flood_stage * 100):0.6f}%",  # 默认值95%
                # 这个值要大于等于cluster.routing.allocation.disk.watermark.high，否则设置不成功，这时候只能搜索，不能加了
            }
        }
        print(body)
        url = f'{self.host}/_cluster/settings'
        resp = requests.put(url, json=body, timeout=5)
        out = resp.json()
        return out

    def set_not_read_only(self):
        body = {'index.blocks.read_only_allow_delete': None}
        url = f'{self.host}/_all/_settings'
        resp = requests.put(url, json=body)
        out = resp.json()
        return out

    def set_max_result_window(self, index, num=5000000):
        self.create_index(index)
        url = f'{self.host}/{index}/_settings'
        body = {"index.max_result_window": num}
        resp = requests.put(url, json=body)
        out = resp.json()
        return out

    def set_field_type(self, index, field, dtype):
        assert dtype in ['geo', 'nest']
        self.create_index(index)
        if dtype == 'geo':
            dtype = 'geo_point'
        elif dtype == 'nest':
            dtype = 'nested'
        url = f'{self.host}/{index}/_mapping'
        body = {"properties": {field: {"type": dtype}}}
        resp = requests.put(url, json=body)
        out = resp.json()  # {"acknowledged": true}
        return out

    def put_mapping(self, index, mapping):
        """字段类型概述
            一级分类	二级分类	具体类型
            核心类型	字符串类型  text,  keyword
            整数类型	integer,long,short,byte
            浮点类型	double,float,half_float,scaled_float
            逻辑类型	boolean
            日期类型	date
            范围类型	range
            二进制类型	binary
            复合类型	数组类型	array
            对象类型	object
            嵌套类型	nested
            地理类型	地理坐标类型	geo_point
            地理地图	geo_shape
            特殊类型	IP类型	ip
            范围类型	completion
            令牌计数类型	token_count
            附件类型	attachment
            抽取类型	percolator
        """
        url = f'{self.host}/{index}/_mapping'
        resp = requests.put(url, json=mapping)
        return resp.json()

    async def async_index_exist(self, index):
        url = f'{self.host}/{index}'
        try:
            async with httpx.AsyncClient() as client:
                resp = await client.get(url, headers=self.headers, timeout=3)
            out = 'error' not in resp.json()
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}   超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  {error}')
            out = None
        return out

    async def async_search(self, index, query):
        url = f'{self.host}/{index}/_doc/_search'
        try:
            async with httpx.AsyncClient() as client:
                resp = await client.post(url, json=query, headers=self.headers, timeout=3)
            out = resp.json()
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}  query={query}  超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  query={query}  {error}')
            out = None
        return out

    def search(self, index, query, timeout=20):
        url = f'{self.host}/{index}/_doc/_search'
        try:
            resp = requests.post(url, json=query, headers=self.headers, timeout=timeout)
            out = resp.json()
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}  query={query}  超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  query={query}  {error}')
            out = None
        return out

    def agg(self, index, query):
        url = f'{self.host}/{index}/_doc/_search'
        try:
            resp = requests.post(url, json=query, headers=self.headers, timeout=3)
            out = resp.json()
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}  query={query}  超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  query={query}  {error}')
            out = None
        return out

    async def async_agg(self, index, query):
        url = f'{self.host}/{index}/_doc/_search'
        try:
            async with httpx.AsyncClient() as client:
                resp = await client.post(url, json=query, headers=self.headers, timeout=3)
            out = resp.json()
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'url={url}  query={query}  超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'url={url}  query={query}  {error}')
            out = None
        return out

    def scroll(self, index, query, wait='10m', timeout=20000):
        if isinstance(query, Query):
            query = query()
        if not self.index_exist(index):
            return []
        return ScrollIter(self.host, index, query, wait=wait, timeout=timeout)

    def scroll_batch(self, index, query, wait='10m', timeout=20000):
        if isinstance(query, Query):
            query = query()
        if not self.index_exist(index):
            return []
        return ScrollBatch(es=self, index=index, query=query, wait=wait, timeout=timeout)

    async def async_del_doc(self, index, doc_id):
        url = f'{self.host}/{index}/_doc/{doc_id}'
        try:
            async with httpx.AsyncClient() as client:
                resp = await client.delete(url, headers=self.headers, timeout=3)
            resp_obj = resp.json()
            logs.rotate.info(resp_obj)
            out = True
            if not out:
                logs.rotate.error(f'del resp={resp_obj}')
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'del url={url}  超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'del url={url}  {error}')
            out = None
        return out

    def del_doc(self, index, doc_id):
        url = f'{self.host}/{index}/_doc/{doc_id}'
        try:
            resp = requests.delete(url, headers=self.headers, timeout=5)
            resp_obj = resp.json()
            logs.rotate.info(resp_obj)
            out = True
            if not out:
                logs.rotate.error(f'del resp={resp_obj}')
        except requests.exceptions.ReadTimeout:
            logs.rotate.error(f'del url={url}  超时错误')
            out = None
        except Exception:
            error = '\n'.join(traceback.format_exception(*sys.exc_info()))
            logs.rotate.error(f'del url={url}  {error}')
            out = None
        return out

    def del_field(self, index, field, query=None):
        url = f'{self.host}/{index}/_doc/_update_by_query'
        if query is None:
            query = Query.exist(field)()
        elif isinstance(query, Query):
            query = query()
        body = {'query': query['query'], 'script': {
            'source': f'ctx._source.remove("{field}");',
            'lang': 'painless',
        }}
        print(body)
        resp = requests.post(url, headers=self.headers, json=body, timeout=5)
        resp_obj = resp.json()
        return resp_obj

    def del_index(self, index):
        url = f'{self.host}/{index}'
        logs.rotate.info(f'url={url}')
        try:
            resp = requests.delete(url, headers=self.headers, timeout=3)
            logs.rotate.info(f'index={index} resp={resp.text}')
            if resp.status_code == 404:
                return {'ok': False, 'detail': f'index="{index}" 不存在'}
            if resp.status_code == 200 and resp.json().get('acknowledged'):
                return {'ok': True, 'detail': f'index="{index}" 已经删除'}
            return {'ok': False, 'detail': resp.text}
        except Exception as e:
            logs.rotate.error(f'error={e}')
            return {'ok': False, 'detail': '未获取ES的返回数据'}

    async def async_cat_indices(self):
        url = f'{self.host}/_cat/indices?format=json'
        async with httpx.AsyncClient() as client:
            resp = await client.get(url, headers=self.headers, timeout=3)
            resp_obj = resp.json()
            for index in resp_obj:
                for k, v in list(index.items()):
                    if '.' in k:
                        index[k.replace('.', '_')] = index.pop(k)  # 前端用 .的话会被当做属性引用，所以替换成下划线
            logs.rotate.info(f'resp={resp_obj}')
            return resp_obj

    def get_bulk_txt(self, docs):
        txts = []
        for i, doc in enumerate(docs):
            for k in ['_index', 'opt']:
                if k not in doc:
                    raise ValueError(f'docs[{i}] 缺少key "{k}"')
            opt = doc['opt']
            assert opt in ['create', 'delete', 'update', 'index'], f'opt错误 "{opt}"'
            if opt != 'delete' and '_source' not in doc:
                raise ValueError(f'docs[{i}] 缺少key "_source"')

            line = {opt: {'_index': doc['_index']}}
            if '_id' in doc:
                line[opt]['_id'] = doc['_id']
            txts.append(json.dumps(line) + '\n')
            if opt in ['index', 'create']:
                txts.append(json.dumps(doc['_source']) + '\n')
            elif opt == 'update':
                txts.append(json.dumps({'doc': doc['_source']}) + '\n')
        txt = ''.join(txts)
        return txt

    async def async_save_bulk(self, docs, is_break=True):
        # 存储一个bulk
        # docs = [{_index: '', _id: '', _source: {}, 'opt': 'update'}]
        # opt可取值: update局部更新   index可以创建也可以全量替换   create智能创建   delete删除
        url = f'{self.host}/_bulk'
        if not docs:
            return
        txt = self.get_bulk_txt(docs)
        async with httpx.AsyncClient() as client:
            resp = await client.post(url, data=txt, headers=self.headers, timeout=30)
        resp_obj = resp.json()
        if resp_obj.get('error'):
            all_ok = False
        elif resp_obj.get('errors'):
            all_ok = False
        else:
            all_ok = True
        if not all_ok:
            if 'items' in resp_obj:
                # 筛选掉 操作成功的item
                resp_obj['items'] = [t for t in resp_obj['items'] if any([info.get('status') != 200 for info in t.values()])]
            logs.brief.error(resp_obj)
            if is_break:
                logs.brief.error('exit()')
                exit()
        return all_ok

    def save_bulk(self, docs, is_break=True):
        # 存储一个bulk
        # docs = [{_index: '', _id: '', _source: {}, 'opt': 'update'}]
        # opt可取值: update局部更新   index可以创建也可以全量替换   create智能创建   delete删除
        url = f'{self.host}/_bulk'
        if not docs:
            return
        txt = self.get_bulk_txt(docs)
        resp = requests.post(url, data=txt, headers=self.headers, timeout=30)
        resp_obj = resp.json()
        if resp_obj.get('error'):
            all_ok = False
        elif resp_obj.get('errors'):
            all_ok = False
        else:
            all_ok = True
        if not all_ok:
            if 'items' in resp_obj:
                # 筛选掉 操作成功的item
                resp_obj['items'] = [t for t in resp_obj['items'] if any([info.get('status') != 200 for info in t.values()])]
            logs.brief.error(resp_obj)
            if is_break:
                logs.brief.error('exit()')
                exit()
        return all_ok

    def save_bulk_all(self, docs, size=5000, desc='save_bulk_all', is_break=True):
        # 通过bulk接口把数据存入ES
        # docs = [{_index: '', _id: '', _source: {}, 'opt': 'update'}]
        # opt可取值: update局部更新   index可以创建也可以全量替换   create智能创建   delete删除
        all_is_ok = True
        for i in tqdm(range(0, len(docs), size), desc=desc):
            all_is_ok = all_is_ok and self.save_bulk(docs[i:i + size], is_break=is_break)
        return all_is_ok

    def backup_save(self, zip_path, *indexs):
        # 备份ES， 存储到文件夹中
        zip_path = os.path.abspath(zip_path)
        os.makedirs(os.path.dirname(zip_path), exist_ok=True)
        if os.path.exists(zip_path):
            raise InfoError(f'{zip_path} 文件已存在')
        index__mapping = {}
        for index in indexs:
            index__mapping[index] = self.get_map(index)
        z = ZipTool(zip_path, 'w')
        z.add('index__mapping.pkl', pickle.dumps(index__mapping))
        query = Query.all(size=1000)  # size 不能太大， 避免requests传递data不支持大数据
        for index in indexs:
            for i, batch in enumerate(tqdm(self.scroll_batch(index, query), desc=f'backup_save: {index}')):
                arcname = f'{index}/{i}.pkl'
                content = pickle.dumps(batch)
                z.add(arcname, content)
        logs.rotate.info(f'已完成备份 {zip_path}')

    def backup_load(self, zip_path, opt='index'):
        # 从备份中恢复数据到ES     opt=update局部更新 / index可以创建也可以全量替换
        f = ZipTool(zip_path, 'r')
        index__ok_ids = {}
        errors = []
        index__mapping = pickle.loads(f.read('index__mapping.pkl'))
        for index, mapping in index__mapping.items():
            logs.rotate.info(f'backup_load index={index}')
            if not self.index_exist(index):
                self.create_index(index)
            resp = self.put_mapping(index, mapping[index]['mappings'])
            if not resp.get('acknowledged'):
                logs.rotate.error(f'index({index}) mapping ERROR')
                logs.rotate.error(str(resp))
                raise InfoError('index mapping ERROR')

        for name in tqdm(f.namelist(), desc='backup_load'):
            if name == 'index__mapping.pkl':
                continue
            batch = pickle.loads(f.read(name))
            for doc in batch:
                doc['opt'] = opt
            try:
                self.save_bulk(batch)
                for doc in batch:
                    index__ok_ids.setdefault(doc['_index'], []).append(doc['_id'])
            except:
                errors.append(name)
        logs.rotate.info(f'已恢复备份 {zip_path}  errors~{len(errors)}')
        return index__ok_ids, errors

    def get_agg_list(self, index, field, query=None, size=100):
        # 如果field值是文本，则需要末尾添加 .keyword    如果是整数则不需要添加
        # return  = [{'doc_count': 数量, 'key': '文本内容'}]
        query = query or Query.all()
        q = Query.aggs({'resp': {'field': field, 'size': size}}, query=Query.extract_query(query))
        resp = self.agg(index, q)
        if 'aggregations' in resp and 'resp' in resp['aggregations'] and 'buckets' in resp['aggregations']['resp']:
            return resp['aggregations']['resp']['buckets']
        else:
            raise InfoError(f'resp={resp}')

    async def async_get_agg_list(self, index, field, query=None, size=100):
        # 如果field值是文本，则需要末尾添加 .keyword    如果是整数则不需要添加
        # return  = [{'doc_count': 数量, 'key': '文本内容'}]
        query = query or Query.all()
        q = Query.aggs({'resp': {'field': field, 'size': size}}, query=Query.extract_query(query))
        resp = await self.async_agg(index, q)
        if 'aggregations' in resp and 'resp' in resp['aggregations'] and 'buckets' in resp['aggregations']['resp']:
            return resp['aggregations']['resp']['buckets']
        else:
            raise InfoError(f'resp={resp}')

    def get_agg_count(self, index, field, query=None):
        # 返回 聚类 bulket 数量
        q = {
            'size': 0,
            'aggs': {
                'Key1': {
                    'terms': {
                        'field': field,
                    },
                },
                "count": {
                    "cardinality": {
                        "field": field
                    }
                }
            }
        }
        if query:
            q['query'] = Query.extract_query(query)
        resp = self.agg(index, q)
        return resp['aggregations']['count']['value']

    def get_map(self, index=''):
        if index:
            url = f'{self.host}/{index}/_mapping'
        else:
            url = f'{self.host}/_mapping'
        resp = requests.get(url)
        out = resp.json()
        return out


if __name__ == '__main__':
    es = EsTool('http://127.0.0.1:9200')
    kv = {
        '_fn': 'bool',
        '_param': {'must': [{'_fn': 'term', '_param': {'key': 'key1.keyword', 'values': 'xxxxx'}}]},
        '_call': {'start': 10}
    }
    # x = Query.json_to_query(kv)
    # pprint(x)

    # print(es.set_disk())
    # print(es.set_not_read_only())
    # print(es.all_index())
    data = [
        {'_index': 'test33', '_source': {'value': 1}, 'opt': 'index'},
        {'_index': 'test33', '_source': {'value': 2}, '_id': 'test2', 'opt': 'index'},
    ]
    es.save_bulk(data)
